#!/usr/bin/perl -w
#############################################################################
## 
## 
## NLANR Multicast Beacon
## http://dast.nlanr.net/Projects/Beacon
## 
## Copyright (c) 1999,2000,2001,2002,  The Board of Trustees of the University of Illinois 
## All Rights Reserved.
## 
## Developed by:                         Kai Chen
##                                     Mitch Kutzko
##                                     Tony Rimovsky
## 
##                      National Center for Supercomputing Applications
##                		University of Illinois at Urbana-Champaign
##                		       http://www.ncsa.uiuc.edu
## 
## Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
## and associated documentation files (the "Software"), to deal in the Software without 
## restriction, including without limitation the rights to use, copy, modify, merge, publish, 
## distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the 
## Software is furnished to do so, subject to the following conditions: 
## 
##   - Redistributions of source code must retain the above copyright notice, this list of 
##     conditions and the following disclaimers. 
## 
##   - Redistributions in binary form must reproduce the above copyright notice, this list 
##     of conditions and the following disclaimers in the documentation and/or other 
##     materials provided with the distribution.
## 
##   - Neither the names of the University of Illinois, NCSA, nor the names of its 
##     contributors may be used to endorse or promote products derived from this 
##     Software without specific prior written permission. 
## 
## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
## EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
## OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
## NONINFRINGEMENT. IN NO EVENT SHALL THE CONTIBUTORS OR 
## COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
## LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
## ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE 
## USE OR OTHER DEALINGS IN THE SOFTWARE.
## 
#############################################################################
##
## What:
##	NLANR Multicast BeaconServer, Server code
##
##	Version 0.8.09 - Last Edited: February 7, 2002, Mitch
##
##	January, 2002  - Mitch Kutzko
##	mitch@dast.nlanr.net | mitch@ncsa.uiuc.edu
##	Feedback to beacon@dast.nlanr.net
##
## Filename and Version:
##	beaconserver_v0.8.09_server.pl
##
## Description:
##	Listens on a socket and generates a list of beacons that 
##	it sees reporting on that socket.  It reports those beacons, along with
##	the amount of delay, jitter, loss, outoforder, and duplicate packets 
##	between the current and remote beacon pairs, sorted by reverse
##	domain name.  Writes the collected data out to output files 
##	"table.out" and "table2.out" every $UPDATEINTERVAL seconds.
##	The data are reported in the order they occur in the packet:  
##	Delay, Jitter, Loss, OOO, Duplicate.  These files are used by 
##	the HTML generator to create the actual web files.
##
##
## Background:
##
##	This script is a Perl version of the v0.63 BeaconServer part 
##	of the NLANR Multicast Beacon, originally written in Java by 
##	Kai Chen.  See http://dast.nlanr.net/Projects/Beacon/.
##
## Required Modules:
## 	The perl modules listed in the 'use' statements at the beginning:
##	"IO::Socket", "IO::Socket::Multicast", and "Socket"
##
##	"TimeDate" will also need to be installed for this script to 
## 	function properly.
##
## Setup:
##	If your perl install is up to date, all you should need to do 
##	is edit the beaconserver_v0.8.5_server.pl file and change the 
##	multicast group and port ($GROUP and $PORT) variables to point 
##	to the desired group and run it.
##
##	In .../beacon/start_beacon.sh for the Beacon *Client*, the 
##	"multicast_port" value is the port the Beacon Clients 
##	talk to each other on.  The "server_port" value is the 
##	port the Beacon Clients talk to the Beacon Server on.  
##	The value of $PORT should be the same as the value of 
##	"server_port" in the start_beacon.sh script for the 
##	Beacon Clients.
##
##
##
## Notes:
##
## In the interest of simplicity and scalability, all this script 
## (beaconserver_v0.8.09_server.pl) does is generates the two output 
## files, "table.out" and "table2.out".  It does nothing else at all, 
## unlike previous version of the BeaconServer.  The HTML files are
## generated from the output files by a separate script cleverly 
## named "beaconserver_v0.8.09_htmlgen.pl".
##
## In the interest of simplicity and scalability, the 
## BeaconServer really has no business pretending to be a web 
## server, and so it doesn't.
##
## The htmlgenerator.pl script creates the HTML files from the table2.out file.
##
## Serving those files to the world is up to you.  We suggest running your favorite web
## server on the same machine as the BeaconServer, and linking the HTML files that the 
## BeaconServer generates into a directory underneath your web server's /docroot.
## Note that no "index.html" file is generated -- Typically, you just create an
## additional link to whichever set of data you want to display first as "index.html".
## For most folks, this means linking "loss.html" to "index.html" in the web directory.
##
## All files generated by these scripts are generated in the local directory where 
## the scripts are run.
##
## "table2.out" is the only output file actually used by this version of the
## BeaconServer -- "table.out" is the original format of the file, and is provided 
## here purely as a convenience for any other applications which may already use 
## that format.
##
## The user the web server runs as needs to have permissions to read the various files
## generated by these scripts.
##
## IMPORTANT NOTE:
## In this version, it is necessary to let the beaconserver.pl script run for 
## one $UPDATEINTERVAL prior to starting the htmlgenerator.pl script, else the 
## beaconserver_v0.8.09_htmlgen.pl script will fall over due to not finding the 
## table2.out file initially.  (This will be made more robust in a future version.)
##
##
## Acknowledgements:
## NLANR would like to specifically acknowledge the much-appreciated assistance of
## ANL's Bob Olson, NDSU's Marty Hoag, and the ACCESS Grid Community in the testing 
## and debugging of this script.
##
##
## This version of the Perl BeaconServer script was written by Mitch 
## Kutzko, NLANR/DAST, based on an earlier version of the script written 
## by Tony Rimovsky and Jon Dugan of NCSA Netdev (netdev@ncsa.uiuc.edu).      
##
## Mitch Kutzko, mitch@ncsa.uiuc.edu, November 2001
##
#########################################################################################

##################################
## Overall housekeeping stuff
##################################

# Set autoflush to true so Perl doesn't wait for a "newline" to print output
$| = 1;


# Generate a tempporary PID file, so when we restart the process, we know what we're whacking.
$perlpidfile = ">perlbeaconserver\.PID";
open(PERLPIDFILE, $perlpidfile) || die ("PERLPIDFILE open failed" );
print PERLPIDFILE "$$";         # $$ is perl for "process ID", so I know what to kill next time.
close (PERLPIDFILE);

##################################
## Required modules listed here
##################################
use IO::Socket;
use IO::Socket::Multicast;

use Socket;

# Verion
$VER = "0.8.09";

# Multicast Replicator IP and port number
$GROUP = "233.4.200.1";	# The group and port number for some NLANR test Beacons. 
$PORT  = "10100";       # Port 10100 is the multicast group that all the test Beacons send their information on.

#$GROUP = "233.2.171.1";# The group and port number for the ANL Access Grid Beacons. 
#$PORT  = "9997";       # Port 9997 is the multicast group that all the ANL Beacons send their information on.

require "ctime.pl";
## Also requires that the 'TimeDate' module be installed, too...

##################################
## Configuration info
##################################

# Max length of incoming datagram packet
$MAXLEN = 65536;  	# Per nothing -- Pulled a power of two out of my a.... the air.    ;-)

$UPDATEINTERVAL	= 120;	# Interval (in seconds) to kick out web update at

$CRBLIMIT	= 6;	# Number of things (-1 for the 0-slot) in Current Reporting Beacon info.
$DATALIMIT	= 7;	# Number of pieces of data per reportED beacon



#########################################################################################
##
## getstats()	- Is handed one Beacon's report packet, and parses it to retrieve
##		information from that particular Beacon.  
##
#########################################################################################
#
#
# get session name - Session name is unique for all entries in this whole packet.  Session name
# is the unique ID for this whole multicast group/channel/space/thingie.
#
#####################################
# The entire packet of information here is broken up 
# at <byte for length of next field><next field>:, and
# the final entry is a "#" to indicate termination of the whole packet.
#
# This grabs the first byte of the packet to snag the length of the session 
# name - "c" is a formatting instruction to get the data being unpacked to an 
# unsigned char (read "byte" - Go figure...), and the
# substr grabs from the zeroth position in var for a single byte.

sub getstats() {


# Read in single Beacon's worth (one UDP socket connection) of data/report info.

# This grabs the next "$len" characters of the packet ($packet), which happens to be the 
# session name.

$packetlth = length($packet);			# How long is the current packet we're working with?

$len = unpack("c", substr($packet, 0, 1));      # grab first byte for length (-unpack- does atoi)

$session = substr($packet, 1, $len);		# first chunk in the packet is the multicast session name
$pointer = $len+1;				# advance ptr to the length byte of the second field in packet

$dataflag = 1;					# assume that there will be some data from the CRB (Currently Reporting Beacon)

# Get Current Reporting Beacon (CRB) info. (Session = \"$session\"\n";

$ii = 0;	# Loop through all <CRBLIMIT> pieces of info about the Currently Reporting Beacon
while ($ii < $CRBLIMIT ) {
	
	# DEBUG for accessgrid02.ncl.ac.uk bad beacon data problem - 8.08
	if ($len > $packetlth) {
		#print "pre: Bad Length - \$len = $len, \$ii = $ii\n";
		$dataflag = 0;		# Bad data.  No biscuit!
		# Just whack this Beacon now and go back for a new packet.  Bogus data here.
		if ($ii > 0) {		# If $ii > 0, we've already assigned crbinfo{$b_name}[$ii] to some value
			delete $crbinfo{$b_name}; # Delete info from the Currently Reporting Beacon array
		
			if (exists $crbinfo{$b_name}) {
				print "CRBINFO for \"$b_name\" not deleted successfully.\n";
			}
		}
		return;					# go back and get another packet -- This one's toast
	} elsif ($pointer > $packetlth) {
		#print "Pre: Bad Pointer - \$pointer = $pointer, \$ii = $ii\n";
		$dataflag = 0;		# Bad data.  No biscuit!
		if ($ii > 0) {		# If $ii > 0, we've already assigned crbinfo{$b_name}[$ii] to some value
			delete $crbinfo{$b_name}; # Delete info from the Currently Reporting Beacon array
		
			if (exists $crbinfo{$b_name}) {
				print "CRBINFO for \"$b_name\" not deleted successfully.\n";
			}
		}
		return;					# go back and get another packet -- This one's toast
	}
	# END DEBUG

	$len =  unpack("c", substr($packet, $pointer, 1)); # get next record length from next length byte
	$pointer++;					# increment pointer to start of next chunk of data

	# DEBUG for accessgrid02.ncl.ac.uk bad beacon data problem
	if ($len > $packetlth) {
		#print "Post: Bad Length - \$len = $len, \$ii = $ii\n";
		$dataflag = 0;		# Bad data.  No biscuit!
		if ($ii > 0) {		# If $ii > 0, we've already assigned crbinfo{$b_name}[$ii] to some value
			delete $crbinfo{$b_name}; # Delete info from the Currently Reporting Beacon array
		
			if (exists $crbinfo{$b_name}) {
				print "CRBINFO for \"$b_name\" not deleted successfully.\n";
			}
		}
		return;					# go back and get another packet -- This one's toast
	} elsif ($pointer > $packetlth) {
		#print "Post: Bad Pointer - \$pointer = $pointer, \$ii = $ii\n";
		$dataflag = 0;		# Bad data.  No biscuit!
		if ($ii > 0) {		# If $ii > 0, we've already assigned crbinfo{$b_name}[$ii] to some value
			delete $crbinfo{$b_name}; # Delete info from the Currently Reporting Beacon array
		
			if (exists $crbinfo{$b_name}) {
				print "CRBINFO for \"$b_name\" not deleted successfully.\n";
			}
		}
		return;					# go back and get another packet -- This one's toast
	}
	# END DEBUG


	$data = unpack("x$pointer A$len", $packet);	# read next $len chars of packet string into "$data"
	#$data = substr($packet, $pointer, $len);

	# assign current piece of info: 0 = CRB name, 1 = ip, 2 = OS, 3 = OS ver, 4 = Arch, 5 = JVM
	if ($ii == 0) {					# First piece is the name of the Reporting Beacon
		$b_name = lc($data);			# Force b_name into lowercase for sorting
        }

	$data =~ s/ /&nbsp;/g;				# Convert spaces to &nbsp's so OS Names don't wrap

	# save data to crbinfo array: 0 = name, 1 = ip, 2 = OS, 3 = OS ver, 4 = Arch, 5 = JVM
	$crbinfo{$b_name}[$ii] = $data;

	# Encountering hashmark ("end of entire packet") here means CRB has no data to report.
	# Encountering a null here means that we're getting bad data from the CRB.
	# In either case, this CRB needs to be dropped from the list.
	if (substr($packet, $pointer+$len, 1) eq "#" || $data eq "") {
		$dataflag = 0;	# Clear the "data to be reported" flag, so we'll skip reporting this beacon

		# Just whack this Beacon now and go back for a new packet.  Bogus data here.
		delete $crbinfo{$b_name}; # Delete info from the Currently Reporting Beacon array

		if (exists $crbinfo{$b_name}) {
			print "CRBINFO for \"$b_name\" not deleted successfully.\n";
		}
		return;					# go back and get another packet -- This one's toast
	}

	$pointer += $len;				# increment pointer past current data to next length 
	$ii++;						# Increment the loop index
}


# Save the CRB name here, because we can't save it till it's actually known.

if ($dataflag) {	# Only add this if the crbinfo hasn't been deleted for missing or bad data
	$crbinfo{$b_name}[0] = $b_name;		# b_name

	# Beacon info doesn't really go this far.  It actually only goes to $CRBLIMIT-1.
	# "$session" isn't technically part of the packet being saved here, but I include it here
	# just to keep it with the CRB info.  
	$crbinfo{$b_name}[$CRBLIMIT] = $session;	

} else {				# dataflag set to 0, bad or mising data for this CRB....
        delete $crbinfo{$b_name};       # Delete info from the Currently Reporting Beacon array
        if (exists $crbinfo{$b_name}) {
                print "CRBINFO for \"$b_name\" not deleted successfully.\n";
        }
        return;                                 # go back and get another packet -- This one's toast
}



# Grab all the length/value pairs for Remote Beacons -- Beacons being reported on by CRB.

$dataflag = 1;	# Assume full data

$state = 0;	# initialize state for decoding packet chunks - $DATALIMIT chunks total/beacon

# loop until the hashmark that marks the end of data, as long as there's actually data to get.
while	(($pointer < $packetlth) && ($dataflag)) {
	$len =  unpack("c", substr($packet, $pointer, 1)); # get record length from length byte

	$pointer++;					# increment pointer to start of next chunk of data

	$data = unpack("x$pointer A$len", $packet);	# skip pointer and read next $len chars into data
	#$data = substr($packet, $pointer, $len);

	if ($data eq "") {	# If any of the data values comes up NULL, drop this Beacon and its data
		$dataflag = 0;	# Clear the "data to be reported" flag, so we'll skip reporting this beacon
	}

	if ($dataflag) {	# There's data for this beacon -- Process it

		# 0 = name, 1 = last reported date, (Both handled above)
		# 2 = average delay, 3 = jitter, 4 = loss, 
		# 5 = out of order, 6 = duplicate datagrams

		if ($state == 0) {				# Remote Beacon Name
			$remote_name = lc($data);		# Force remote_name into lowercase for sorting

		} elsif ($state == 1) {		# Epoch time: Last time Beacon heard from Remote Beacon
			# The time value we get here is three chars too long, because right now it's
			# coming from a client written in Java.  Javatime != epochtime!!!!  (Since it was
			# originally going to a Java server, it didn't matter.  Now it does, so "fix" it here.)
			#print DEBUGFILE "\"$data\" - length given $len\n";
			#$stats{$b_name}{$remote_name}[$state] = substr($data,0,$len-3);

		} elsif ($state == 4) {	# Loss
			$stats{$b_name}{$remote_name}[$state] = $data * 100;

		} else {

			$stats{$b_name}{$remote_name}[$state] = $data;

		}
	}

	$pointer += $len;		# increment pointer past data
	$state++;			# increment to the next type of reporting data
	if ($state == $DATALIMIT) {	# Wrap back around to start of reporting data types.

		if ($dataflag) {	# Only save this name if we've got valid data for this Beacon
			# Save the remote_name, too....  (Gotta do it down here, though, after everything else...)
			$stats{$b_name}{$remote_name}[0] = $remote_name;
		}

		$state = 0;	# reset the state counter so we can go get another chunk o' data!
	}
}

return();
}



#########################################################################################
##
## tableoutput()	- Loops over the array of all just-reported Beacon data and builds a web page.
##
#########################################################################################

sub tableoutput {

# open the temporary outfile - Where we'll be writing HTML data off to prior to swapping it into the real file....
$datetemp = &ctime(time);				# Get the current time

@across = @down = sort {$a cmp $b} keys %stats;			# set up matrix for output page & snag total # beacons
$totalbeacons = $#down + 1;				# Length of @down array

$linecount = 0;	 	# Actual lines of output to the web page for display.
foreach $rb (@down) {	# this loop runs over the array of sending beacon names

	# Now print this row for Receivers

	print TABLEFILE "$linecount ";			# line number for Bob's table.out file
	print TABLEFILE2 "$linecount ";			# line number for Bob's table.out file

	# Save the stats now......
	# 0 = name, 1 = last reported date, 2 = average delay 3 = jitter, 
	# 4 = loss, 5 = out of order, 6 = duplicate datagrams

	foreach $sb (@across) {
		# Now print the tuples for Bob's PHP script for table.out
		# 	Swapped "exists" for "defined" for Bob (Perl 5.001) on 1/22/02 - Mitch
		#if (exists $stats{$rb}{$sb}[2]) {	# Don't assign if UNDEF - If good data, write stats
		if (defined $stats{$rb}{$sb}[2]) {	# Don't assign if UNDEF - If good data, write stats
			print TABLEFILE "$stats{$rb}{$sb}[2]/";
			print TABLEFILE "$stats{$rb}{$sb}[3]/";
			print TABLEFILE "$stats{$rb}{$sb}[4]/";
			print TABLEFILE "$stats{$rb}{$sb}[5]/";
			print TABLEFILE "$stats{$rb}{$sb}[6] ";

			print TABLEFILE2 "$stats{$rb}{$sb}[2]/";
			print TABLEFILE2 "$stats{$rb}{$sb}[3]/";
			print TABLEFILE2 "$stats{$rb}{$sb}[4]/";
			print TABLEFILE2 "$stats{$rb}{$sb}[5]/";
			print TABLEFILE2 "$stats{$rb}{$sb}[6] ";

		} else {				# No data
			print TABLEFILE "NA ";
			print TABLEFILE2 "-/-/-/-/- ";
		}

	}
	
	print TABLEFILE "\n";			# Newline at end of line of tuples.
	print TABLEFILE2 "\n";			# Newline at end of line of tuples.

	$linecount++;	# web page line counter
}

return();
}



#########################################################################################
##
## crboutput()  - Loops over the array of CRBinfo and build table of reporting Beacons
##
#########################################################################################

sub crboutput {

$linecount = 0;         				# Actual lines of output to the web page for display.
foreach $crb (sort keys %crbinfo) {			# Loops over currently reporting Beacons

	# Print out the Beacon-specific info here.  Goes to $CRBLIMIT+1 to 
	# get $session, which is tacked onto the end of crbinfo just to have 
	# it here even though it's not part of the packet.
	#
	# CRB info: 0 = CRB name, 1 = ip, 2 = OS, 3 = OS ver, 4 = Arch, 5 = JVM

	print TABLEFILE "beacon $linecount\n"; # Print line number for each beacon to Bob's table.out file
	print TABLEFILE2 "beacon $linecount\n"; # Print line number for each beacon to Bob's table.out file

	for ($cc=0; $cc<$CRBLIMIT; $cc++) {
		# CRB info: 0 = CRB name, 1 = ip, 2 = OS, 3 = OS ver, 4 = Arch, 5 = JVM
		if ($cc == 0) {$tablestring = "name";}
		if ($cc == 1) {$tablestring = "ip";}
		if ($cc == 2) {$tablestring = "osName";}
		if ($cc == 3) {$tablestring = "osVersion";}
		if ($cc == 4) {$tablestring = "osArch";}
		if ($cc == 5) {$tablestring = "vmVersion";}
		print TABLEFILE "$tablestring $crbinfo{$crb}[$cc]\n";
		print TABLEFILE2 "$tablestring $crbinfo{$crb}[$cc]\n";
	}

        $linecount++;   # web page line counter
}

return();
}



#########################################################################################
##
## tableoutsetup()  - Starts the table.out file for Bob's PHP script
##
#########################################################################################

sub tableoutsetup {

# open the table.out file  - Where we'll be writing Bob's parsable output data off to....
$tablefile = ">table\.TEMP";
open(TABLEFILE, $tablefile) || die ("TABLEFILE open failed" );
$tablefile2 = ">table2\.TEMP";
open(TABLEFILE2, $tablefile2) || die ("TABLEFILE2 open failed" );

$timegrab = time;					# Grab epoch time
$datetemp = &ctime(time);				# Get the current time

#@down = sort keys %stats;				# snag total # beacons
@down = sort {$a cmp $b} keys %stats;			# set up matrix for output page & snag total # beacons
$totalbeacons = $#down + 1;				# Length of @down array

print TABLEFILE "$VER" . "_updated $timegrab $datetemp";	# "1003952479 Wed Oct 24 14:41:19 2001"
print TABLEFILE "nbeacons $totalbeacons\n";
#print TABLEFILE2 "updated $timegrab $datetemp";   	# "1003952479 Wed Oct 24 14:41:19 2001"
print TABLEFILE2 "$VER" . "_updated $timegrab $datetemp";	# "1003952479 Wed Oct 24 14:41:19 2001"
print TABLEFILE2 "nbeacons $totalbeacons\n";
print TABLEFILE2 "group $GROUP port $PORT\n";		# include multicast group and port

$linecount = 0;         				# Actual lines of output to the web page for display.
foreach $crb (sort keys %crbinfo) {			# Loops over currently reporting Beacons

	# "beacon 0 Fordlab@DisplayUofM 136.8.33.105" --  beacon number, name, and ip....
	print TABLEFILE "beacon $linecount $crbinfo{$crb}[0] $crbinfo{$crb}[1]\n";
	print TABLEFILE2 "beacon $linecount $crbinfo{$crb}[0] $crbinfo{$crb}[1]\n";
	$linecount++;
}

return();
}






################################################################################
##
## Beginning of main routine
##
################################################################################


########################################
#
# This section sends a query to the beacon server asking
# for the report for a multicast group and the port the group is using
# 
########################################
#
# Open the socket.
#
# SOCKET is the file handle it returns, PF_INET is the Protocol Family 
# domain (as opposed to the Arpanet Family domain), SOCK_DGRAM is how we want the 
# data - DGRAM means chunks are ok -- SOCK_STREAM would mean that we want the 
# incoming data reassembled into a stream for use.  getprotobyname("udp")
# tells Perl to go look in /etc/protocols and find out what port "udp" is bound 
# to, and use that port	number.

my $recvsock = IO::Socket::Multicast->new(LocalPort=>$PORT);
$recvsock->mcast_add($GROUP) || die "Couldn't set group: $!\n";



$datetemp = &ctime(time);	# Get the current time
print "NLANR BeaconServer v$VER (Perl) - Listening to Multicast Group $GROUP:$PORT on $datetemp";

$mark = time;					# Grab the current time so we know when to update the web page

while (1) {					# Main loop

	next unless $recvsock->recv($packet, $MAXLEN); 	# Only work with complete packet of data

	getstats();	# Process the current packet-worth of data
	$packet = "";	# clear the packet so we know we're not recycling data - Otherwise we reprocess stale data!!

	if (time - $mark >= $UPDATEINTERVAL) {	# Update the web page every $UPDATEINTERVAL seconds
		$datetemp = &ctime(time);	# Get the current time
		$mark = time;

		tableoutsetup();		# Start the table.out file output
		tableoutput();			# Output the data FROM the Beacons

		crboutput();			# Get the beacon metadata

		close(TABLEFILE);		# Close Bob's table.out file now that it's finished
		close(TABLEFILE2);		# Close Bob's table.out file now that it's finished

		# Swap the real files into the real names.  "*Whump*!"
		rename "table\.TEMP", "table\.out";
		rename "table2\.TEMP", "table2\.out";

		undef %stats;	# Clear the vars used for most recent set of web pages so we're not recycling data.
		undef %crbinfo;

	}
}

exit(0);
